import numpy as np
import pandas as pd
from sklearn.neighbors import KNeighborsClassifier
from collections import Counter
import time
import matplotlib.pyplot as plt

start = time.time()
# Load the table data for all NEOs observed with TCS.
tcs_path = "/home/gabriel/Gabriel/Astrophysics/BD4SB/TCS-MUSCAT2/CSV_files/KNN_training_set.csv"
tcs = pd.read_csv(tcs_path, encoding = "utf-16")


"""Training sets"""
three_colors_training = tcs.index[(tcs.gmr_MedColor.notna()) & (tcs.rmi_MedColor.notna()) & (tcs.imz_MedColor.notna())
                    & ~(tcs.Training_set == "-")].values
# The case when all three colors are known
rmi_imz_training = tcs.index[(tcs.imz_MedColor.notna()) & (tcs.rmi_MedColor.notna())
                             & ~(tcs.Training_set == "-")].values
# The case when only rmi and imz colors are known.
gmr_rmi_training = tcs.index[(tcs.gmr_MedColor.notna()) & (tcs.rmi_MedColor.notna())
                             & ~(tcs.Training_set == "-")].values
# The case when only gmr and rmi colors are known.


"""Sample sets"""
three_colors_sample = tcs.index[(tcs.gmr_MedColor.notna()) & (tcs.rmi_MedColor.notna()) & (tcs.imz_MedColor.notna())
                                & (tcs.Training_set == "-")]
# The sample set of objects with three colors and without taxonomy for which we'll apply the KNN.
rmi_imz_sample = tcs.index[(tcs.rmi_MedColor.notna()) & (tcs.imz_MedColor.notna()) & (tcs.gmr_MedColor.isna())
                           & (tcs.Training_set == "-")].values
# The sample set of objects with rmi and imz colors and without taxonomy for the KNN.
gmr_rmi_sample = tcs.index[(tcs.gmr_MedColor.notna()) & (tcs.rmi_MedColor.notna()) & (tcs.imz_MedColor.isna())
                           & (tcs.Training_set == "-")].values
# The sample set of objects with gmr and rmi colors and without taxonomy for the KNN.


def knn_algorithm(object_desig, size):  # KNN algorithm for one object

    KNN = KNeighborsClassifier(n_neighbors = 3, metric = "minkowski", p = 2)  # Initialize a KNN algorithm
    object_index = tcs.index[tcs.MainDesig == object_desig].values.item()  # Object's index
    taxonomy_table = pd.Series(dtype = str)

    if object_index in three_colors_sample:  # The case when all three colors are known
        training_colors = np.array(tcs.loc[three_colors_training, ["gmr_MedColor", "rmi_MedColor", "imz_MedColor"]],
                                   dtype = float)  # Colors
        training_tax = tcs.Training_set.loc[three_colors_training]  # The corresponding taxonomies
        model = KNN.fit(training_colors, training_tax)  # The training model

        monte_carlo = np.random.normal(loc = np.array(tcs.loc[object_index, ["gmr_MedColor", "rmi_MedColor",
                "imz_MedColor"]], dtype = float), scale = np.array(tcs.loc[tcs.MainDesig == test_obj,
                ["gmr_ColorErr", "rmi_ColorErr", "imz_ColorErr"]]), size = (size, 3))

        """Here we generate a normal distribution for each pair (mean = color, std = color_err) in order to account for 
         magnitude errors. Then, for each case we apply the KNN and chose the taxonomy which is in majority."""

        # Here we store the assigned taxonomy for each color set, who's been generated in "monte_carlo".
        taxonomy_object = pd.Series(dtype = str)
        # Apply the KNN for all colors in monte-carlo
        taxonomy_object.loc[0] = model.predict(monte_carlo.reshape(size, 3))
        # Append at each step the taxonomy to taxonomy_object list
        taxonomy_table = taxonomy_table.append(taxonomy_object, ignore_index = True)

    elif object_index in rmi_imz_sample:  # The case when only rmi and imz are available
        training_colors = np.array(tcs.loc[rmi_imz_training, ["rmi_MedColor", "imz_MedColor"]], dtype = float)
        training_tax = tcs.Training_set.loc[rmi_imz_training]

        model = KNN.fit(training_colors, training_tax)  # Training model (with two colors)

        monte_carlo = np.random.normal(loc = np.array(tcs.loc[object_index, ["rmi_MedColor", "imz_MedColor"]],
                dtype = float), scale = np.array(tcs.loc[object_index, ["rmi_ColorErr", "imz_ColorErr"]],
                dtype = float), size = (size, 2))

        taxonomy_object = pd.Series(dtype = str)
        taxonomy_object.loc[0] = model.predict(monte_carlo.reshape(size, 2))
        taxonomy_table = taxonomy_table.append(taxonomy_object, ignore_index = True)

    elif object_index in gmr_rmi_sample:  # The case when only gmr and rmi are available
        training_colors = np.array(tcs.loc[gmr_rmi_training, ["gmr_MedColor", "rmi_MedColor"]], dtype = float)
        training_tax = tcs.Training_set.loc[gmr_rmi_training]

        model = KNN.fit(training_colors, training_tax)

        monte_carlo = np.random.normal(loc = np.array(tcs.loc[object_index, ["gmr_MedColor", "rmi_MedColor"]],
                dtype = float), scale = np.array(tcs.loc[object_index, ["gmr_ColorErr", "rmi_ColorErr"]],
                dtype = float), size = (size, 2))

        taxonomy_object = pd.Series(dtype = str)
        taxonomy_object.loc[0] = model.predict(monte_carlo.reshape(size, 2))
        taxonomy_table = taxonomy_table.append(taxonomy_object, ignore_index = True)

    predicted_tax = Counter(taxonomy_table[0])  # Count how many times each taxonomy is predicted.

    return predicted_tax.most_common()


def knn_output():
    """Run this function if you wish to apply the KNN for all objects with colors and without taxonomy."""

    KNN_output_table = pd.DataFrame(columns = ["MainDesig", "No_colors", "Predict_tax1", "Predict_probab1",
                                               "Predict_tax2", "Predict_probab2"])
    KNN_sample_index = np.sort(np.concatenate([three_colors_sample, rmi_imz_sample, gmr_rmi_sample]))
    # Row's index of all objects without taxonomies.

    for j in KNN_sample_index:
        KNN_output_object = pd.DataFrame(columns = ["MainDesig", "No_colors", "Predict_tax1", "Predict_probab1",
                                               "Predict_tax2", "Predict_probab2"])
        KNN_output_object.loc[0, "MainDesig"] = tcs.MainDesig.loc[j]

        if j in three_colors_sample:
            KNN_output_object.loc[0, "No_colors"] = 3
        else:
            KNN_output_object.loc[0, "No_colors"] = 2

        knn_output = knn_algorithm(tcs.MainDesig.loc[j], size = 10000)

        KNN_output_object.loc[0, "Predict_tax1"] = knn_output[0][0]
        KNN_output_object.loc[0, "Predict_probab1"] = knn_output[0][1] * 1e-4

        if KNN_output_object.loc[0, "Predict_probab1"] < 0.7:
            KNN_output_object.loc[0, "Predict_tax2"] = knn_output[1][0]
            KNN_output_object.loc[0, "Predict_probab2"] = knn_output[1][1] * 1e-4

        KNN_output_table = KNN_output_table.append(KNN_output_object, ignore_index = True)

    return KNN_output_table


# Apply KNN for one object
test_obj = "153957"
tcs_knn = knn_algorithm(test_obj, size = 10000)

# Uncomment the following lines if you want to apply KNN for all bodies
# knn_out = knn_output()
# knn_export_path = "/home/gabriel/Gabriel/Astrophysics/BD4SB/TCS-MUSCAT2/CSV_files/KNN3_Output.csv"
# knn_out.to_csv(knn_export_path, encoding = "utf-16", columns = knn_out.columns, index = False)
stop = time.time() - start  # The running time (in seconds).

