#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Wed Apr 11 15:33:33 2021

@author: marcel
"""
from astropy.io import fits
import numpy as np
import pandas as pd
import glob
import os


def make_night_log(foldername):
    os.chdir(os.getenv("HOME"))
    from pyraf import iraf
    iraf.unlearn('imcombine')
    
    imags = []
    filenames = glob.glob(foldername+"*.fits")
    for filename in filenames:
        with fits.open(filename) as hdul:
#        hdul.info()
            hdr = hdul[0].header
#       print(repr(hdr))
            imags.append([hdr['FRAMEID'], hdr['OBJECT'], hdr['FILTER'],
                          hdr['DATE-OBS'], hdr['EXP-STRT'], hdr['EXPTIME'], 
                          hdr['RA'], hdr['DEC'], hdr['AIRMASS']])
            if (os.path.isdir(foldername + hdr['OBJECT'] + '/' +  hdr['FILTER'])==False):
                if (os.path.isdir(foldername + hdr['OBJECT'])==False):
                    os.system('mkdir ' + foldername + hdr['OBJECT'])
                os.system('mkdir ' + foldername + hdr['OBJECT'] + '/' +  hdr['FILTER'])
            #os.system('mv ' + filename + ' ' + foldername + hdr['OBJECT'] +
            #          '/' +  hdr['FILTER'] + '/')
            iraf.imcopy(filename[:-5]+'[11:1014,51:1014]', foldername + hdr['OBJECT'] +
            '/' +  hdr['FILTER'] + '/' + filename.split('/')[-1], Stdout=1)
            os.system('rm ' + filename)
            
    
    df2 = pd.DataFrame(imags,columns=('FRAMEID', 'OBJECT', 'FILTER',
                                      'DATE-OBS', 'EXP-STRT', 'EXPTIME',
                                      'RA','DEC', 'AIRMASS'))
    df2.to_csv(path_or_buf=foldername + 'NightLog.csv',index=False)
    bands = np.unique(df2.FILTER)
    return bands





def combine_bias(foldername, bands):

    os.chdir(os.getenv("HOME"))
    from pyraf import iraf

    iraf.unlearn('imcombine')
    iraf.imcombine.combine = 'median'
    iraf.imcombine.scale = 'none'

    for filt in bands:
        iraf.cd(foldername + 'DARK/' + filt)
        os.system('rm -f master_bias* biaslist')
        os.system('ls MCT*.fits > biaslist')
        iraf.imcombine('@biaslist ', 'master_bias', Stdout=1)
        os.system('mv master_bias.fits ../master_bias_' + filt + '.fits')
        os.system('rm biaslist')
    return

def apply_bias(foldername, objlist, bands):
    
    objlist.append('FLAT')
    os.chdir(os.getenv("HOME"))
    from pyraf import iraf
    iraf.unlearn('imarith')
    
    for myobj in objlist:
        for filt in bands:
            iraf.cd(foldername + myobj + '/' + filt)
            os.system('rm -f rembias_*.fits imaglist')            
            os.system('ls MCT*.fits > imaglist')
            iraf.imarith('@imaglist ', '-', foldername + 'DARK/master_bias_' + filt + '.fits', 'rembias_//@imaglist')
    return


def combine_flat(foldername, bands):

    os.chdir(os.getenv("HOME"))
    from pyraf import iraf

    iraf.unlearn('imcombine')
    iraf.imcombine.combine = 'median'
    iraf.imcombine.scale = 'none'
    iraf.unlearn('imarith')

    for filt in bands:
        iraf.cd(foldername + 'FLAT/' + filt)
        os.system('rm -f master_flat* flatslist')
        os.system('ls rembias_*.fits > flatslist')
        iraf.imcombine('@flatslist ', 'master_flat', Stdout=1)
        mymed = iraf.imstat('master_flat',fields='midpt',format='no',Stdout=1)
        iraf.imarith('master_flat','/',float(mymed[0]),'norm_master_flat')
        os.system('mv norm_master_flat.fits ../master_flat_' + filt + '.fits')
        os.system('rm flatslist')
    return

def apply_flat(foldername, objlist, bands):
    
    os.chdir(os.getenv("HOME"))
    from pyraf import iraf
    iraf.unlearn('imarith')
    
    for myobj in objlist:
        for filt in bands:
            iraf.cd(foldername + myobj + '/' + filt)
            os.system('rm -f proc_*.fits imaglist')            
            os.system('ls rembias_*.fits > imaglist')
            os.system('split --lines 100 imaglist imaglist')
            filenames = glob.glob("imaglista*")
            #print(filenames)
            for imaglist in filenames:
                iraf.imarith('@'+imaglist+'', '/', foldername + 'FLAT/master_flat_' + filt + '.fits', 'proc_//@'+imaglist)
            os.system('rm -f rembias_*.fits MCT*.fits') 
            os.system('rm -f imaglist*') 
    return


def preparefolder(foldername):
    
    # Make night log
    bands = make_night_log(foldername)
    # Make master bias     
    combine_bias(foldername,bands)
    # Apply bias to flat fields
    apply_bias(foldername, ['FLAT'], bands)
    # Make master flat
    combine_flat(foldername, bands)
    
    #Prepare list of objects
    objects = os.listdir(foldername)
    objects.remove('FLAT')
    objects.remove('DARK')
    objects.remove('NightLog.csv')

    apply_bias(foldername,objects, bands)
    apply_flat(foldername,objects, bands)


    return

#preparefolder('/scratch1/MUSCAT/180801/')
#preparefolder('/scratch1/MUSCAT/180924/')
#preparefolder('/scratch1/MUSCAT/181009/')
#preparefolder('/scratch/2I/20191001/')
#preparefolder('/scratch1/MUSCAT/191003/')
preparefolder('/scratch/2I/ReducedData/20191121/')


